<?php

namespace Drupal\oidc\OpenidConnectRealm;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Url;
use Drupal\oidc\JsonWebTokens;

/**
 * Interface for an OpenID Connect realm.
 */
interface OpenidConnectRealmInterface extends PluginInspectionInterface {

  /**
   * Indicates whether the realm is enabled.
   *
   * @return bool
   *   The realm status.
   */
  public function isEnabled();

  /**
   * Get the login URL.
   *
   * @param string $state
   *   A unique state to identify this login attempt.
   * @param Url $redirect_url
   *   The login redirect URL.
   *
   * @return \Drupal\Core\Url
   *   The login URL.
   */
  public function getLoginUrl($state, Url $redirect_url);

  /**
   * Get the tokens.
   *
   * @param string $state
   *   The state to identify the login attempt.
   * @param string $code
   *   The authorization code.
   *
   * @return \Drupal\oidc\JsonWebTokens
   *   The retrieved tokens.
   *
   * @throws \RuntimeException
   */
  public function getJsonWebTokens($state, $code);

  /**
   * Get the logout URL.
   *
   * @param \Drupal\oidc\JsonWebTokens $tokens
   *   The JSON web tokens.
   * @param string $state
   *   A unique state to identify this logout request.
   * @param string $redirect_url
   *   The logout redirect URL.
   *
   * @return \Drupal\Core\Url
   *   The logout URL.
   */
  public function getLogoutUrl(JsonWebTokens $tokens, $state, Url $redirect_url);

  /**
   * Fetch the JSON web keys set and add any missing keys to the local storage.
   *
   * @return bool
   *   Boolean indicating if the update succeeded.
   */
  public function updateJwks();

  /**
   * Clear the JSON web keys set storage.
   */
  public function clearJwks();

  /**
   * Get the display name format.
   *
   * @return string
   *   The display name in tokens.
   */
  public function getDisplayNameFormat();

}
